/**
  *  \file server/router.cc
  *  \brief PCC2 Router
  *
  *  This provides session management/routing. It starts Server
  *  processes, manages their lifetime, and routes requests to them.
  *
  *  This module is Unix-specific.
  */

#include <list>
#include <memory>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <time.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include "cpluslib/types.h"
#include "cpluslib/string.h"
#include "cpluslib/format.h"
#include "server/session.h"
#include "server/request.h"
#include "server/config.h"
#include "server/net.h"
#include "filenotify.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "9999";
static string_t arg_server = "./c2server";
static int arg_timeout = 10000;
static int arg_virgintimeout = 60;
static int arg_sessionlimit = 10;
static bool arg_newsessionswin = false;
static bool arg_filenotify = true;

static uint32_t session_counter;
static std::list<RouterSession*> sessions;

static void handleTimeouts();

static void
logCommand(const string_t& verb, const char* command)
{
    if (*command) {
        printf("> '%s %s'\n", verb.c_str(), command);
    } else {
        printf("> '%s'\n", verb.c_str());
    }
}

static RouterSession*
handleNewSessionRequest(string_t arg, const char*& msg)
{
    /* Create new session object */
    std::auto_ptr<RouterSession> n(new RouterSession(arg, ++session_counter));

    /* Check and resolve conflicts */
    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        if (int result = n->checkConflict(**i, arg_newsessionswin)) {
            /* We have a conflict */
            if (result > 0) {
                /* New sessions wins */
                delete *i;
                i = sessions.erase(i);
            } else {
                /* Old sessions wins */
                msg = "453 Conflicting session already active\n";
                return 0;
            }
        } else {
            ++i;
        }
    }

    /* Check for overload */
    if (int(sessions.size()) >= arg_sessionlimit) {
        handleTimeouts();
    }
    if (int(sessions.size()) >= arg_sessionlimit) {
        msg = "451 Too many sessions\n";
        return 0;
    }

    /* Start the session */
    if (n->start(arg_server.c_str())) {
        /* Session started successfully */
        RouterSession* nn = n.release();
        sessions.push_back(nn);
        return nn;
    } else {
        /* Session start failed */
        return 0;
    }
}

/** Handle LIST request.
    \param fd File descriptor to write output to */
static void
handleListRequest(int fd)
{
    char buf[100];
    sprintf(buf, "200 OK, %d sessions\n", sessions.size());
    writeStr(fd, buf);
    time_t now = time(0);
    for (std::list<RouterSession*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
        writeStr(fd,
                 format("%10d  %6d %6ds  %c%c %s\n")
                 << (*i)->getSessionId()
                 << (*i)->getPid()
                 << int32_t(difftime(now, (*i)->getLastAccessTime()))
                 << ((*i)->wasUsed() ? '.' : 'V')
                 << ((*i)->wasModified() ? '.' : 'S')
                 << (*i)->getName());
    }
}

/** Handle INFO request.
    \param fd File descriptor to write output to
    \param session  Session id */
static void
handleInfoRequest(int fd, uint32_t session)
{
    for (std::list<RouterSession*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
        if ((*i)->getSessionId() == session) {
            writeStr(fd, "200 OK\n");
            (*i)->writeArgs(fd);
            return;
        }
    }
    writeStr(fd, "452 Session timed out\n");
}

/** Read and ignore a S (talk) request. */
static void
readTalkRequest(Request& req, int fd, uint32_t session)
{
    /* Read one line */
    char* command;
    while ((command = req.readLine()) == 0) {
        if (req.readMore(fd) != Request::sSuccess) {
            // writeStr(fd, "400 Bad Request\n");
            return;
        }
    }

    /* Log it */
    printf("invalid %d [?]: '%s'\n", int(session), command);
    if (isPOST(command)) {
        while (1) {
            char* line = req.readLine();
            if (line == 0) {
                /* No complete line; try reading some more data */
                Request::State st = req.readMore(fd);
                if (st != Request::sSuccess
                    && !(st == Request::sTooLong && req.shift()))
                {
                    break;
                }
            } else if (strcmp(line, ".") == 0) {
                break;
            } else {
                /* ignore */
            }
        }
    }
}

/** Handle S (talk) request. */
static void
handleTalkRequest(Request& req, int fd, uint32_t session)
{
    /* Locate the session */
    for (std::list<RouterSession*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
        if ((*i)->getSessionId() == session) {
            /* Found it. Execute request. */
            RouterSession* s = *i;
            s->executeRequest(req, fd);
            if (!s->isActive()) {
                /* This command caused the session to terminate. An
                   answer for the client has already been produced.
                   So all we need to do is to delete the session. */
                delete *i;
                sessions.erase(i);
            }
            return;
        }
    }

    readTalkRequest(req, fd, session);
    writeStr(fd, "452 Session timed out\n");
}

class GroupCommand {
 public:
    virtual bool handle(std::list<RouterSession*>::iterator it) = 0;
    virtual void report(int fd, uint32_t numFailures, uint32_t numSuccesses) = 0;

    void run(int fd, const char* command);
};


void
GroupCommand::run(int fd, const char* command)
{
    /* Collect Ids */
    std::vector<uint32_t> ids;
    int numFailures = 0;

    /* Syntax check */
    if (command[0] == 0 || (command[0] != '-' && command[strspn(command, "0123456789")] != 0)) {
        writeStr(fd, "400 Bad request\n");
        return;
    }

    /* Do it */
    uint32_t desiredId = strtoul(command, 0, 0);
    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        /* Fetch item */
        std::list<RouterSession*>::iterator ii = i;
        ++i;

        /* Want this? */
        uint32_t thisId = (*ii)->getSessionId();
        bool wantThis = (command[0] == '-' ? (*ii)->checkConflict(command, false, true) < 0 : thisId == desiredId);

        /* Process it */
        if (wantThis) {
            if (handle(ii)) {
                ids.push_back(thisId);
            } else {
                ++numFailures;
            }

            /* Did this operation cause the session to go away? */
            if (!(*ii)->isActive()) {
                delete *ii;
                sessions.erase(ii);
            }
        }
    }

    /* Report */
    report(fd, numFailures, ids.size());
    if (command[0] == '-') {
        for (std::vector<int32_t>::size_type i = 0; i < ids.size(); ++i) {
            writeStr(fd, format("%d\n", ids[i]));
        }
    }
}


class CloseCommand : public GroupCommand {
 public:
    bool handle(std::list<RouterSession*>::iterator it)
        {
            (*it)->stop();
            return true;
        }

    void report(int fd, uint32_t /*numFailures*/, uint32_t numSuccesses)
        {
            writeStr(fd, format("200 OK, %d sessions closed\n", numSuccesses));
        }
};

class RestartCommand : public GroupCommand {
 public:
    bool handle(std::list<RouterSession*>::iterator it)
        {
            return (*it)->restart(arg_server.c_str());
        }

    void report(int fd, uint32_t numFailures, uint32_t numSuccesses)
        {
            if (numSuccesses) {
                writeStr(fd, format("200 OK\n"));
            } else if (numFailures > 0) {
                writeStr(fd, "500 Restart failed\n");
            } else {
                writeStr(fd, "452 Session timed out\n");
            }
        }
};

class SaveCommand : public GroupCommand {
    bool notify;
 public:
    SaveCommand(bool notify)
        : notify(notify)
        { }
    bool handle(std::list<RouterSession*>::iterator it)
        {
            return (*it)->save(notify);
        }

    void report(int fd, uint32_t /*numFailures*/, uint32_t numSuccesses)
        {
            writeStr(fd, format("200 OK, %d sessions saved\n", numSuccesses));
        }
};

static void
handleRequest(int fd)
{
    Request req;

    /* Read command */
    char* command;
    while ((command = req.readLine()) == 0) {
        if (req.readMore(fd) != Request::sSuccess) {
            writeStr(fd, "400 Bad request\n");
            return;
        }
    }

    /* Find command verb */
    string_t verb;
    while (*command != 0 && *command != ' ' && *command != '\t') {
        verb += toupper(uint8_t(*command++));
    }
    while (*command == ' ' || *command == '\t') {
        ++command;
    }

    /* Process request */
    if (verb == "LIST") {
        /* @q LIST (Router Command)
           List active sessions.
           Active sessions will be reported as a table:
           - session Id
           - PID
           - idle time in seconds
           - "," if the session was used, "V" if it is still virgin (indicates a browser without JavaScript)
           - "." if the session was modified, "S" if it was saved
           - command line parameters of the session */
        logCommand(verb, command);
        handleListRequest(fd);
    } else if (verb == "INFO") {
        /* @q INFO id:RouterSession (Router Command)
           Get information about a session.
           Reports all command line parameters, one per line.
           @err 452 Session timed out (no such session) */
        logCommand(verb, command);
        if (command[strspn(command, "0123456789")] != 0) {
            writeStr(fd, "400 Bad request\n");
        } else {
            handleInfoRequest(fd, strtoul(command, 0, 0));
        }
    } else if (verb == "S") {
        /* @q S id:RouterSession (Router Command)
           Talk to a session.
           The actual request (e.g. "GET ...") follows on a line after the command.
           For POST requests, the POST body follows after the request.
           The response is equal to the response reported by the server.
           @err 452 Session timed out (no such session)
           @err 400 Bad request (%id invalid, or request cannot be read)
           @err 500 Invalid answer from server */
        if (command[strspn(command, "0123456789")] != 0) {
            /* FIXME: should still read complete request */
            writeStr(fd, "400 Bad request\n");
        } else {
            handleTalkRequest(req, fd, strtoul(command, 0, 0));
        }
    } else if (verb == "CLOSE") {
        /* @q CLOSE {id:RouterSession | -flag} (Router Command)
           When given an %id, closes that session.

           When given a %flag, closes all sessions that conflict with that flag.
           The session Ids of all closed sessions are reported with the response. */
        logCommand(verb, command);
        CloseCommand().run(fd, command);
    } else if (verb == "RESTART") {
        /* @q RESTART {id:RouterSession | -flag} (Router Command)
           When given an %id, restarts that session.

           When given a %flag, restarts all sessions that conflict with that flag.
           The session Ids of all restarted sessions are reported with the response.

           @err 500 Restart failed (some sessions could not be restarted)
           @err 452 Session timed out */
        logCommand(verb, command);
        RestartCommand().run(fd, command);
    } else if (verb == "SAVE") {
        /* @q SAVE {id:RouterSession | -flag} (Router Command)
           When given an %id, saves that session.

           When given a %flag, saves all sessions that conflict with that flag.
           The session Ids of all saved sessions are reported with the response.

           @uses FORGET (File Command) */
        logCommand(verb, command);
        SaveCommand(true).run(fd, command);
    } else if (verb == "SAVENN") {
        /* @q SAVENN {id:RouterSession | -flag} (Router Command)
           When given an %id, saves that session.

           When given a %flag, saves all sessions that conflict with that flag.
           The session Ids of all saved sessions are reported with the response.

           Unlike {SAVE (Router Command)}, this command does not notify the {File (Service)|File Service}. */
        logCommand(verb, command);
        SaveCommand(false).run(fd, command);
    } else if (verb == "NEW") {
        /* @q NEW args:Str... (Router Command)
           Start new session.
           The new session identifier will be reported as NN:{@type RouterSession} in a "201 NN Created" response.

           @err 403 Forbidden (generail failure)
           @err 451 Too many sessions
           @err 453 Conflicting session already active
           @rettype RouterSession */
        logCommand(verb, command);
        const char* msg = "403 Forbidden\n";
        RouterSession* s = handleNewSessionRequest(command, msg);
        if (s != 0) {
            writeStr(fd, format("201 %d Created\n", s->getSessionId()));
        } else {
            writeStr(fd, msg);
        }
    } else if (verb == "CONFIG") {
        /* @q CONFIG (Router Command)
           Report the configuration. */
        writeStr(fd, "200 OK\n");
        writeStr(fd, format("Router.Host=%s\n"
                            "Router.Port=%s\n"
                            "Router.Timeout=%d\n"
                            "Router.VirginTimeout=%d\n"
                            "Router.MaxSessions=%d\n"
                            "Router.NewSessionsWin=%d\n"
                            "Router.FileNotify=%d\n")
                 << arg_listen_host
                 << arg_listen_port
                 << arg_timeout
                 << arg_virgintimeout
                 << arg_sessionlimit
                 << arg_newsessionswin
                 << arg_filenotify);
    } else {
        writeStr(fd, "400 Bad request\n");
    }
}

static void
handleTimeouts()
{
    time_t now = time(0);

    std::list<RouterSession*>::iterator i = sessions.begin();
    while (i != sessions.end()) {
        int timeout = (*i)->wasUsed() ? arg_timeout : arg_virgintimeout;
        if (difftime(now, (*i)->getLastAccessTime()) > timeout) {
            printf("c2router: session %d timed out\n", int((*i)->getSessionId()));
            delete *i;
            i = sessions.erase(i);
        } else {
            ++i;
        }
    }
}

static void
handleRouterConfig(string_t key, string_t value)
{
    if (key == "ROUTER.HOST") {
        /* @q Router.Host:Str (Config)
           Listen address for Router service. */
        arg_listen_host = value;
    } else if (key == "ROUTER.PORT") {
        /* @q Router.Port:Int (Config)
           Port number for Router service. */
        arg_listen_port = value;
    } else if (key == "ROUTER.SERVER") {
        /* @q Router.Server:Str (Config)
           File name of %c2server binary. */
        arg_server = value;
    } else if (key == "ROUTER.TIMEOUT") {
        /* @q Router.Timeout:Int (Config)
           Session timeout in seconds.
           A session will be terminated if it has not been accessed within this time.
        */
        stringValue2(value, arg_timeout);
    } else if (key == "ROUTER.VIRGINTIMEOUT") {
        /* @q Router.VirginTimeout:Int (Config)
           Session timeout in seconds for virgin (unaccessed) sessions.
           A session will be terminated if it has not been accessed within this time directly after creation.
           This happens when a user starts a session but their browser has trouble with the JavaScript. */
        stringValue2(value, arg_virgintimeout);
    } else if (key == "ROUTER.MAXSESSIONS") {
        /* @q Router.MaxSessions:Int (Config)
           Maximum number of concurrent sessions. */
        stringValue2(value, arg_sessionlimit);
    } else if (key == "ROUTER.NEWSESSIONSWIN") {
        /* @q Router.NewSessionsWin:Str (Config)
           Determines behaviour when two conflicting sessions are started ("-W" and "-R" flags).
           If "y" or "1", new sessions that conflict with old ones cause the old ones to terminate.
           If "n" or "0", the new session will be refused. */
        arg_newsessionswin = isTrue(value);
    } else if (key == "ROUTER.FILENOTIFY") {
        /* @q Router.FileNotify:Str (Config)
           If "y" or "1", the {SAVE (Router Command)|SAVE} command will notify the {File (Service)|file server}. */
        arg_filenotify = isTrue(value);
    }
    handleFileServerConfig("FILE", key, value);
}

int main()
{
    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Load configuration */
    loadConfiguration(handleRouterConfig);

    /* File server notify */
    if (arg_filenotify && !initNotifyFileServer()) {
        fprintf(stderr, "unable to initialize file server notification\n");
        return 1;
    }

    /* Where do we want to listen? */
    int sock = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (sock < 0) {
        fprintf(stderr, "unable to bind socket\n");
        return 1;
    }

    /* OK, main loop */
    printf("c2router: listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(sock, F_SETFD, FD_CLOEXEC);
    listen(sock, 10);
    while (1) {
        int fd = accept(sock, 0, 0);
        if (fd < 0) {
            perror("accept");
            break;
        }
        fcntl(fd, F_SETFD, FD_CLOEXEC);
        // sleep(1);
        handleRequest(fd);
        close(fd);
        handleTimeouts();
    }
    close(sock);
}
